// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file StdMpi.h */

#ifndef AMDIS_STDMPI_H
#define AMDIS_STDMPI_H

#include <map>
#include <mpi.h>
#include "MeshStructure.h"
#include "parallel/InteriorBoundary.h"

namespace AMDiS {

  int intSizeOf(std::map<const DegreeOfFreedom*, DegreeOfFreedom> &data)
  {
    return data.size() * 2;
  }

  int intSizeOf(std::vector<MeshStructure> &data)
  {
    int s = 0;
    for (unsigned int i = 0; i < data.size(); i++)
      s += data[i].getCode().size() + 2;

    return s;
  }

  int intSizeOf(std::vector<WorldVector<double> > &data)
  {
    return data.size() * Global::getGeo(WORLD);
  }

  int intSizeOf(std::vector<int> &data)
  {
    return data.size();
  }

  int intSizeOf(std::vector<double> &data)
  {
    return data.size();
  }
  
  int intSizeOf(std::vector<std::pair<int, int> > &data)
  {
    return data.size() * 2;
  }

  int intSizeOf(std::vector<AtomicBoundary> &data)
  {
    return data.size() * 3;
  }

  int intSizeOf(std::vector<BoundaryObject> &data)
  {
    return data.size() * 5;
  }

  int intSizeOf(std::vector<const DegreeOfFreedom*> &data)
  {
    return data.size();
  }

  void makeBuf(std::map<const DegreeOfFreedom*, DegreeOfFreedom> &data, int* buf)
  {
    int i = 0;
    for (std::map<const DegreeOfFreedom*, DegreeOfFreedom>::iterator it = data.begin();
	 it != data.end(); ++it) {
      buf[i++] = *(it->first);
      buf[i++] = it->second;
    }
  }
  
  void makeFromBuf(std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &data, 
		   int *buf, int bufSize)
  {    
    if (bufSize == 0)
      return;

    TEST_EXIT(bufSize % 2 == 0)("This should not happen!\n");

    data.clear();
    data.reserve(bufSize / 2);

    for (int i = 0; i < (bufSize / 2); i++)
      data.push_back(std::make_pair(buf[i * 2], buf[i * 2 + 1]));
  }

  void makeBuf(std::vector<MeshStructure> &data, unsigned long int *buf)
  {    
    int pos = 0;
    for (unsigned int i = 0; i < data.size(); i++) {
      buf[pos++] = data[i].getCode().size();
      buf[pos++] = data[i].getNumElements();

      for (unsigned int j = 0; j < data[i].getCode().size(); j++)
	buf[pos++] = data[i].getCode()[j];
    }
  }

  void makeFromBuf(std::vector<MeshStructure> &data, 
		   unsigned long int *buf, int bufSize)
  {
    int pos = 0;

    while (pos < bufSize) {       
      int codeSize = buf[pos++];
      int nElements = buf[pos++];
      std::vector<unsigned long int> code;
      code.resize(codeSize);
      for (int i = 0; i < codeSize; i++)
	code[i] = buf[pos++];

      MeshStructure meshCode;
      meshCode.init(code, nElements);

      data.push_back(meshCode);
    }	
  }

  void makeBuf(std::vector<WorldVector<double> > &data, double *buf)
  {
    int dimOfWorld = Global::getGeo(WORLD);
    int pos = 0;
    for (unsigned int i = 0; i < data.size(); i++)
      for (int j = 0; j < dimOfWorld; j++)
	buf[pos++] = data[i][j];
  }

  void makeFromBuf(std::vector<WorldVector<double> > &data, double *buf, int bufSize)
  {
    int dimOfWorld = Global::getGeo(WORLD);
    TEST_EXIT(bufSize % Global::getGeo(WORLD) == 0)("This should not happen!\n");

    int pos = 0;
    data.resize(bufSize / Global::getGeo(WORLD));
    for (unsigned int i = 0; i < data.size(); i++)
      for (int j = 0; j < dimOfWorld; j++)
	data[i][j] = buf[pos++];
  }

  void makeBuf(std::vector<int> &data, int *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++)
      buf[i] = data[i];
  }

  void makeFromBuf(std::vector<int> &data, int *buf, int bufSize)
  {
    data.resize(bufSize);

    for (int i = 0; i < bufSize; i++)
      data[i] = buf[i];
  }

  void makeBuf(std::vector<double> &data, double *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++)
      buf[i] = data[i];
  }

  void makeFromBuf(std::vector<double> &data, double *buf, int bufSize)
  {
    data.resize(bufSize);

    for (int i = 0; i < bufSize; i++)
      data[i] = buf[i];
  }

  void makeBuf(std::vector<std::pair<int, int> > &data, int *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++) {
      buf[i * 2] = data[i].first;
      buf[i * 2 + 1] = data[i].second;
    }
  }

  void makeBuf(std::vector<AtomicBoundary> &data, int *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++) {
      buf[i * 3] = data[i].rankObj.elIndex;
      buf[i * 3 + 1] = data[i].rankObj.subObj;
      buf[i * 3 + 2] = data[i].rankObj.ithObj;
    }
  }

  void makeFromBuf(std::vector<AtomicBoundary> &data, int *buf, int bufSize)
  {
    if (bufSize == 0)
      return;

    TEST_EXIT(bufSize % 3 == 0)("This should not happen!\n");    

    data.resize(bufSize / 3);
    for (int i = 0; i < bufSize / 3; i++) {
      data[i].rankObj.elIndex = buf[i * 3];
      data[i].rankObj.subObj = static_cast<GeoIndex>(buf[i * 3 + 1]);
      data[i].rankObj.ithObj = buf[i * 3 + 2];
    }
  }

  void makeBuf(std::vector<BoundaryObject> &data, int *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++) {
      buf[i * 5] = data[i].elIndex;
      buf[i * 5 + 1] = data[i].elType;
      buf[i * 5 + 2] = data[i].subObj;
      buf[i * 5 + 3] = data[i].ithObj;
      buf[i * 5 + 4] = data[i].reverseMode;
    }
  }

  void makeFromBuf(std::vector<BoundaryObject> &data, int *buf, int bufSize)
  {
    if (bufSize == 0)
      return;

    TEST_EXIT(bufSize % 5 == 0)("This should not happen!\n");    

    data.resize(bufSize / 5);
    for (int i = 0; i < bufSize / 5; i++) {
      data[i].elIndex = buf[i * 5];
      data[i].elType = buf[i * 5 + 1];
      data[i].subObj = static_cast<GeoIndex>(buf[i * 5 + 2]);
      data[i].ithObj = buf[i * 5 + 3];
      data[i].reverseMode = static_cast<bool>(buf[i * 5 + 4]);
    }
  }

  template<typename SendT, typename RecvT=SendT>
  class StdMpi 
  {
  public:
    StdMpi(MPI::Intracomm &comm, bool b = true) :
      mpiComm(comm),
      commPrepared(true),
      exchangeDataSize(b)
    {}

    void clear()
    {
      sendData.clear();
      recvData.clear();
      sendDataSize.clear();
      recvDataSize.clear();

      commPrepared = true;
    }

    void send(int toRank, SendT& data)
    {
      FUNCNAME("StdMpi::send()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      sendData[toRank] = data;
      sendDataSize[toRank] = intSizeOf(data);
    }

    void send(std::map<int, SendT>& data)
    {
      FUNCNAME("StdMpi::send()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      for (typename std::map<int, SendT>::iterator it = data.begin(); 
	   it != data.end(); ++it) {
	sendData[it->first] = it->second;
	sendDataSize[it->first] = intSizeOf(it->second);
      }
    }

    std::map<int, SendT>& getSendData()
    {
      return sendData;
    }

    SendT& getSendData(int rank)
    {
      return sendData[rank];
    }

    void updateSendDataSize()
    {
      for (typename std::map<int, SendT>::iterator it = sendData.begin(); 
	   it != sendData.end(); ++it)
	sendDataSize[it->first] = intSizeOf(it->second);
    }

    void recv(int fromRank, int size = -1)
    {
      FUNCNAME("StdMpi::recv()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      recvDataSize[fromRank] = size;
    }

    void recvFromAll(int size = -1)
    {
      FUNCNAME("StdMpi::recvFromAll()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is no prepared!\n");

      for (int i = 0; i < mpiComm.Get_size(); i++)
	recvDataSize[i] = size;
    }

    template<class T>
    void recv(std::map<int, T> &fromRanks)
    {
      FUNCNAME("StdMpi::recv()");
      
      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      for (typename std::map<int, T>::iterator it = fromRanks.begin();
	   it != fromRanks.end(); ++it)
	recvDataSize[it->first] = (exchangeDataSize ? -1 : intSizeOf(it->second));
    }

    std::map<int, RecvT>& getRecvData()
    {
      return recvData;
    }

    RecvT& getRecvData(int rank)
    {
      return recvData[rank];
    }

    void commDataSize()
    {
      MPI::Request request[sendData.size() + recvDataSize.size()];
      std::vector<int> sendBuffers, recvBuffers;
      int requestCounter = 0;

      int i = 0;
      for (typename std::map<int, int>::iterator sendIt = sendDataSize.begin();
	   sendIt != sendDataSize.end(); ++sendIt) {
	sendBuffers.push_back(sendIt->second);
	request[requestCounter++] =
	  mpiComm.Isend(&(sendBuffers[i]), 1, MPI_INT, sendIt->first, 0);
	i++;
      }

      for (std::map<int, int>::iterator recvIt = recvDataSize.begin();
	   recvIt != recvDataSize.end(); ++recvIt)
	request[requestCounter++] =
	  mpiComm.Irecv(&(recvIt->second), 1, MPI_INT, recvIt->first, 0);
      
      MPI::Request::Waitall(requestCounter, request);      
    }

    template<class T>
    void startCommunication(MPI_Datatype m)
    {
      FUNCNAME("StdMpi::startCommunication()");

      TEST_EXIT_DBG(commPrepared)("Communication is not prepared!\n");

      if (exchangeDataSize)
	commDataSize();

      MPI::Request request[sendData.size() + recvDataSize.size()];
      int requestCounter = 0;
      std::vector<T*> sendBuffers, recvBuffers;

      for (typename std::map<int, SendT>::iterator sendIt = sendData.begin();
	   sendIt != sendData.end(); ++sendIt) {
	int bufferSize = sendDataSize[sendIt->first];
	T* buf = new T[bufferSize];
	makeBuf(sendIt->second, buf);

  	request[requestCounter++] = 
  	  mpiComm.Isend(buf, bufferSize, m, sendIt->first, 0);

	sendBuffers.push_back(buf);
      }

      for (std::map<int, int>::iterator recvIt = recvDataSize.begin();
	   recvIt != recvDataSize.end(); ++recvIt) {
	T* buf = new T[recvIt->second];

  	request[requestCounter++] =
  	  mpiComm.Irecv(buf, recvIt->second, m, recvIt->first, 0);
	
	recvBuffers.push_back(buf);	
      }

      MPI::Request::Waitall(requestCounter, request);

      for (int i = 0; i < static_cast<int>(sendBuffers.size()); i++)
	delete [] sendBuffers[i];
      sendBuffers.clear();

      int i = 0;
      for (std::map<int, int>::iterator recvIt = recvDataSize.begin();
	   recvIt != recvDataSize.end(); ++recvIt) {

	makeFromBuf(recvData[recvIt->first], recvBuffers[i], recvIt->second);
	delete [] recvBuffers[i];
	i++;
      }      

      commPrepared = false;
    }

  protected:
    ///
    MPI::Intracomm mpiComm;

    ///
    std::map<int, SendT> sendData;

    ///
    std::map<int, RecvT> recvData;

    std::map<int, int> sendDataSize;

    std::map<int, int> recvDataSize;

    bool commPrepared;

    bool exchangeDataSize;
  };

}

#endif
